# Test the main properties of the Chebyshov polynomials.
# The computations are performed in the algebra of the polynomials in one variable.

# Define and test cheb_T and cheb_U.
# These two functions return the same results as chebyshev_T and chebyshev_U, respectively.

def myvar(letter):
   return PolynomialRing(QQ, letter).gens()[0]

def T_coef(n, k):
   if n == 0 and k == 0:
      return 1
   else:
      return  n * ((-1) ** k) * (2 ** (n - 2 * k - 1)) * factorial(n - k - 1) / (factorial(k) * factorial(n - 2 * k))

def cheb_T(n, a):
   return sum([T_coef(n, k) * (a ** (n - 2 * k)) for k in range(floor(n / 2) + 1)])

def U_coef(n, k):
   return ((-1) ** k) * (2 ** (n - 2 * k)) * binomial(n - k, k)

def cheb_U(n, a):
   return sum([U_coef(n, k) * (a ** (n - 2 * k)) for k in range(floor(n / 2) + 1)])

def test_init_T(t):
   return (cheb_T(0, t) == t ** 0) and (cheb_T(1, t) == t)

def test_recur_T(n, t):
   return cheb_T(n, t) == 2 * t * cheb_T(n - 1, t) - cheb_T(n - 2, t)

def test_init_U(t):
   return (cheb_U(0, t) == t ** 0) and (cheb_U(1, t) == 2 * t)

def test_recur_U(n, t):
   return cheb_U(n, t) == 2 * t * cheb_U(n - 1, t) - cheb_U(n - 2, t)

def test_explicit_T(n, t):
   return cheb_T(n, t) == chebyshev_T(n, t)

def test_explicit_U(n, t):
   return cheb_U(n, t) == chebyshev_U(n, t)

def big_test_cheb_T_pol(nmax):
   b = myvar('b')
   result1 = test_init_T(b)
   result2 = all(test_recur_T(n, b) for n in range(2, nmax + 1))
   result3 = all(test_explicit_T(n, b) for n in range(nmax + 1))
   return result1 and result2 and result3

def big_test_cheb_U_pol(nmax):
   b = myvar('b')
   result1 = test_init_U(b)
   result2 = all(test_recur_U(n, b) for n in range(2, nmax + 1))
   result3 = all(test_explicit_U(n, b) for n in range(nmax + 1))
   return result1 and result2 and result3

# Define and test chebyshev_V and chebyshev_W

def cheb_V(n, a):
   coef = 2 ^ (2 * n) / binomial(2 * n, n)
   return coef * jacobi_P(n, -1/2, 1/2, a)

def cheb_W(n, a):
   coef = 2 ^ (2 * n) / binomial(2 * n, n)
   return coef * jacobi_P(n, 1/2, -1/2, a)

def chebyshev_V(n, a):
   return chebyshev_U(n, a) - chebyshev_U(n - 1, a) if n >= 1 else (a ** 0)

def chebyshev_W(n, a):
   return chebyshev_U(n, a) + chebyshev_U(n - 1, a) if n >= 1 else (a ** 0)

def test_init_V(t):
   return (chebyshev_V(0, t) == t ** 0) and (chebyshev_V(1, t) == 2 * t - 1)

def test_recur_V(n, t):
   return chebyshev_V(n, t) == 2 * t * chebyshev_V(n - 1, t) - chebyshev_V(n - 2, t)

def test_init_W(t):
   return (chebyshev_W(0, t) == t ** 0) and (chebyshev_W(1, t) == 2 * t + 1)

def test_recur_W(n, t):
   return chebyshev_W(n, t) == 2 * t * chebyshev_W(n - 1, t) - chebyshev_W(n - 2, t)

def test_V(n, t):
   return chebyshev_V(n, t) == cheb_V(n, t)

def test_W(n, t):
   return chebyshev_W(n, t) == cheb_W(n, t)

def big_test_cheb_V_pol(nmax):
   b = myvar('b')
   result1 = test_init_V(b)
   result2 = all(test_recur_V(n, b) for n in range(2, nmax + 1))
   result3 = all(test_V(n, b) for n in range(0, nmax + 1))
   return result1 and result2 and result3

def big_test_cheb_W_pol(nmax):
   b = myvar('b')
   result1 = test_init_W(b)
   result2 = all(test_recur_W(n, b) for n in range(2, nmax + 1))
   result3 = all(test_W(n, b) for n in range(0, nmax + 1))
   return result1 and result2 and result3

# Monomials via T or U

def coef_mon_via_cheb_T(n, k):
   c = binomial(n, k)
   return c if 2 * k < n else c / 2

def test_mon_via_cheb_T(n, t):
   lhs = t ^ n
   nhalf = n // 2
   rhs = sum(coef_mon_via_cheb_T(n, k) * 2 * cheb_T(n - 2 * k, t / 2) for k in range(nhalf + 1))
   return lhs == rhs

def element_Catalan_triangle(n, k):
   numer = factorial(n + k) * (n - k + 1)
   denom = factorial(k) * factorial(n + 1)
   return numer / denom

def coef_mon_via_cheb_U(n, k):
   return element_Catalan_triangle(n - k, k)

def test_mon_via_cheb_U(n, t):
   lhs = t ^ n
   nhalf = n // 2
   rhs = sum(coef_mon_via_cheb_U(n, k) * cheb_U(n - 2 * k, t / 2) for k in range(nhalf + 1))
   return lhs == rhs

def big_test_mon_via_cheb_T_pol(nmax):
   b = myvar('b')
   result = all(test_mon_via_cheb_T(n, b) for n in range(nmax + 1))
   return result

def big_test_mon_via_cheb_U_pol(nmax):
   b = myvar('b')
   result = all(test_mon_via_cheb_U(n, b) for n in range(nmax + 1))
   return result

# Main properties

def test_main_property_T(n, t):
   lhs = 2 * chebyshev_T(n, (t + 1 / t) / 2)
   rhs = t ^ n + 1 / t ^ n
   return lhs == rhs

def test_main_property_U(n, t):
   lhs = (t - 1 / t) * chebyshev_U(n, (t + 1 / t) / 2)
   rhs = t ^ (n + 1) - t ^ (- n - 1)
   return lhs == rhs

def test_main_property_V(n, t):
   lhs = (t + 1 / t) * chebyshev_V(n, (t ^ 2 + 1 / t ^ 2) / 2)
   rhs = t ^ (2 * n + 1) + t ^ (- 2 * n - 1)
   return lhs == rhs

def test_main_property_W(n, t):
   lhs = (t - 1 / t) * chebyshev_W(n, (t ^ 2 + 1 / t ^ 2) / 2)
   rhs = t ^ (2 * n + 1) - t ^ (- 2 * n - 1)
   return lhs == rhs

def big_test_main_property_cheb_TUVW_pol(nmax):
   b = myvar('b')
   res = [0] * 4
   res[0] = all(test_main_property_T(n, b) for n in range(nmax + 1))
   res[1] = all(test_main_property_U(n, b) for n in range(nmax + 1))
   res[2] = all(test_main_property_V(n, b) for n in range(nmax + 1))
   res[3] = all(test_main_property_W(n, b) for n in range(nmax + 1))
   return all(res)

# Duplication properties

def test_duplication_cheb_T(n, t):
   lhs = chebyshev_T(2 * n, t / 2)
   rhs = chebyshev_T(n, (t ^ 2 - 2) / 2)
   return lhs == rhs

def test_duplication_cheb_U(n, t):
   lhs = chebyshev_U(2 * n + 1, t / 2)
   rhs = t * chebyshev_U(n, (t ^ 2 - 2) / 2)
   return lhs == rhs

def test_duplication_cheb_V(n, t):
   lhs = 2 * chebyshev_T(2 * n + 1, t / 2)
   rhs = t * chebyshev_V(n, (t ^ 2 - 2) / 2)
   return lhs == rhs

def test_duplication_cheb_W(n, t):
   lhs = chebyshev_U(2 * n, t / 2)
   rhs = chebyshev_W(n, (t ^ 2 - 2) / 2)
   return lhs == rhs

def big_test_duplication_cheb_TUVW_pol(nmax):
   b = myvar('b')
   res = [0] * 4
   res[0] = all(test_duplication_cheb_T(n, b) for n in range(nmax + 1))
   res[1] = all(test_duplication_cheb_U(n, b) for n in range(nmax + 1))
   res[2] = all(test_duplication_cheb_V(n, b) for n in range(nmax + 1))
   res[3] = all(test_duplication_cheb_W(n, b) for n in range(nmax + 1))
   return all(res)

def test_cheb_V_via_cheb_W(n, t):
   lhs = chebyshev_V(n, - t)
   rhs = ((-1) ** n) * chebyshev_W(n, t)
   return lhs == rhs

def big_test_cheb_V_via_cheb_W_pol(nmax):
   b = myvar('b')
   return all(test_cheb_V_via_cheb_W(n, b) for n in range(nmax + 1))

import time

def all_tests_cheb_pol(nmax):
   t = time.time()
   res = [0] * 9
   res[0] = big_test_cheb_T_pol(nmax)
   res[1] = big_test_cheb_U_pol(nmax)
   res[2] = big_test_cheb_V_pol(nmax)
   res[3] = big_test_cheb_W_pol(nmax)
   res[4] = big_test_mon_via_cheb_T_pol(nmax)
   res[5] = big_test_mon_via_cheb_U_pol(nmax)
   res[6] = big_test_main_property_cheb_TUVW_pol(nmax)
   res[7] = big_test_duplication_cheb_TUVW_pol(nmax)
   res[8] = big_test_cheb_V_via_cheb_W_pol(nmax)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

all_tests_cheb_pol(60)

